/* sqUnixMemory -- allocation routines for Squeak's object memory
 *
 */


#include <sq.h>


/* whether to allocate by mmap() on /dev/zero or by malloc().  The
   former is less portable, but it allows the allocated region to grow
   and shrink in place, making -memory irrelevant */

#define MEMORY_USE_MMAP   /* this should be moved into a config file */

#ifdef  MEMORY_USE_MMAP

#include <stdio.h>
#include <unistd.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <assert.h>

/* amount of the memory map to reserve for Squeak. This is the maximum
   amount of memory that Squeak will be able to use, so making it
   large is nice.  However, there are OS-dependent limits on how large
   it can be.  Furthermore, larger values reduce the amount of memory
   available to malloc() and to other uses.  */
#define MEMORY_TO_USE (1024*1024*500)   /* 500 megs */


static int devZeroFd = -1;    /* fd for /dev/zero */
static char *heap = NULL;  /* location of the heap */
static int heapSize = 0;  /* amount of space currently mmap-ed */



/* return the granularity at which to run mmap() */
static int mmapGranularity() 
{
  return getpagesize();
}


/* return the first byte past the end of the heap */
static void *heapLimit() 
{
  return heap + heapSize;
}


/* adjust the size of the heap to at least the requested size */
static void adjustHeapSize(int newHeapSize) 
{
  /* enforce a little sanity */
  if(newHeapSize > MEMORY_TO_USE)
    newHeapSize = MEMORY_TO_USE;
  if(newHeapSize < 0)
    newHeapSize = 0;

  
  /* first, round up newHeapSize so that the heap will end an
     mmapGranularity() boundary */
  {
    char *realEnd;
    int newHeapSizeToUse;
  
    realEnd = (char *)
      (((int)heap + newHeapSize + (mmapGranularity() - 1)) /
       mmapGranularity() * mmapGranularity());

    newHeapSizeToUse = realEnd - heap;
    
    assert(newHeapSizeToUse >= newHeapSize);
    newHeapSize = newHeapSizeToUse;

    assert(newHeapSize >= 0);
    assert(newHeapSize <= MEMORY_TO_USE);
    assert(((int)heap + newHeapSize) % mmapGranularity() == 0);
  }

  
  if(newHeapSize == heapSize) {
    /* no change in size */
    return;
  }
  
  if(newHeapSize < heapSize) {
    /* heap is shrinking */
    void *result;
    
    /* first, munmap pages past the end of the new heap */
    munmap(heap + newHeapSize, MEMORY_TO_USE - newHeapSize);
    heapSize = newHeapSize;

    
    /* now remap them, so that the heap area remains contiguous. */
    /* (This won't actually waste memory on most Unices.) */
    result = mmap(heap+newHeapSize, MEMORY_TO_USE - newHeapSize,
		  PROT_READ|PROT_WRITE, MAP_FIXED|MAP_PRIVATE,
		  devZeroFd, newHeapSize);
    if(result == MAP_FAILED) {
      perror("mmap");
    }
  }
  else {
    /* heap is growing */
    void *result;
    result = mmap(heap+newHeapSize, newHeapSize-heapSize,
		  PROT_READ|PROT_WRITE, MAP_FIXED|MAP_PRIVATE,
		  devZeroFd, newHeapSize);
    if(result == MAP_FAILED) {
      perror("mmap");
      fprintf(stderr, "failed to extend the heap...\n");
      return;
    }
    
    heapSize = newHeapSize ;
  }
}



void * sqAllocateMemory(int minHeapSize, int desiredHeapSize)  {
  /* sanity checks */
  if(heap != NULL) {
    fprintf(stderr, "sqAllocateMemory called twice!\n");
    exit(1);
  }
  if(desiredHeapSize > MEMORY_TO_USE) {
    fprintf(stderr, "requested %d memory for heap!  Giving up.\n", desiredHeapSize);
    return NULL;
  }

  /* open /dev/zero */
  devZeroFd = open("/dev/zero", O_RDWR);
  if(devZeroFd < 0) {
    perror("open(\"/dev/zero\")");
    return NULL;
  }

  /* mmap() the entire MEMORY_TO_USE */
  heap = mmap(NULL, MEMORY_TO_USE,
	      PROT_READ|PROT_WRITE, MAP_PRIVATE,
	      devZeroFd, 0);
  if(heap == MAP_FAILED) {
    perror("mmap");
    heap = NULL;
    return NULL;
  }
  heapSize = MEMORY_TO_USE;
  
  /* adjust the size of the mapping */
  adjustHeapSize(desiredHeapSize);
  if(heapSize == 0) {
    return NULL;
  }


  /* all done */
  return heap;
}


void * sqGrowMemoryBy(void *oldLimit, int delta) {
#if 1
  printf("growing by %d...\n", delta);
#endif
  

  adjustHeapSize((char *)oldLimit - heap + delta);

#if 1
  printf("new heap size is %d\n", heapSize);
#endif
  
  return heapLimit();
}


void * sqShrinkMemoryBy(void *oldLimit, int delta) {
#if 1
  printf("shrinking by %d...\n", delta);
#endif
  
  adjustHeapSize((char *)oldLimit - heap - delta);
  
#if 1
  printf("new heap size is %d\n", heapSize);
#endif
  return heapLimit();
}

int sqMemoryExtraBytesLeft(int includingSwap) {
  return MEMORY_TO_USE - heapSize;
}


#else


void * sqAllocateMemory(int minHeapSize, int desiredHeapSize) {
  malloc(desiredHeapSize);
}

void * sqGrowMemoryBy(void *oldLimit, int delta) {
  return oldLimit;
}

void * sqShrinkMemoryBy(void *oldLimit, int delta) {
  return oldLimit;
}

int sqMemoryExtraBytesLeft(int includingSwap) {
  return 0;
}


#endif
